import weakref, threading, atexit

# This allows to define the 'reusable' decorator.
# You can try for instance:
#
# @reusable
# class C(object):
#     def __init__(self, t):
#         self.t = t
#         print t
#     def __del__(self):
#         print 'del', self.t
#
# c = C(3)
# c = C(3)
# c = C(4)
# c = C(4)
# c = None
#
# When decorated this way, if its constructor is invoked twice
# with the same arguments, the same object will be found in
# cache and returned.

class ReusePool(object):
    def __init__(self, cls):
        self.cls = cls
        self.pool = {}
    def get(self, *args, **kwargs):
        argdesc = args, tuple(kwargs.items())
        if argdesc in self.pool:
            obj = self.pool[argdesc]
        else:
            obj = self.cls(*args, **kwargs)
            self.pool[argdesc] = obj
        return obj

def reusable(cls):
    pool = ReusePool(cls)
    # we decorate a class but we return a function:
    # what seems to be a call to the class constructor
    # will actually be a call to this function.
    def func(*args, **kwargs):
        return pool.get(*args, **kwargs)
    return func





import socket
from fcntl import fcntl, F_GETFD, F_SETFD, FD_CLOEXEC

def udp_server_socket(port):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    # set close-on-exec flag (subprocesses should not inherit it)
    fcntl(s.fileno(), F_SETFD, fcntl(s.fileno(), F_GETFD) | FD_CLOEXEC)
    s.bind(('', port))
    return s



import os, subprocess
from walt.common.version import __version__
from walt.common.tools import restart

# use DEBUG_WITH_TESTPYPI=1 <command>
# to debug the update feature using testpypi.python.org instead
# of the regular pypi repository.

def check_auto_update(server, package_name):
    updated = False
    remote_version = server.get_remote_version()
    if remote_version != int(__version__):
        print('Auto-updating software to match server API...')
        if os.getenv('DEBUG_WITH_TESTPYPI'):
            repo_option = '-i https://testpypi.python.org/simple'
        else:
            repo_option = ''
        cmd = 'sudo pip install %(repo_option)s --upgrade "%(package_name)s==%(remote_version)d"' % \
                        dict(repo_option = repo_option,
                             remote_version = remote_version,
                             package_name = package_name)
        try:
            if subprocess.call(cmd, shell=True) == 0:
                updated = True
        except BaseException as e:
            print('Issue: ' + str(e))
        if not updated:
            print('WARNING!! software update failed. Trying to continue...')
    if updated:
        restart()
    return updated




import inspect
from walt.common.devices import nodes, switches
from walt.common.devices.nodes import *
from walt.common.devices.switches import *

DEVICES = []

def get_node_cls_from_model(model):
    for cls in DEVICES:
        if cls.WALT_TYPE == 'node' and cls.MODEL_NAME == model:
            return cls
    return None     # not found or not a node

def get_device_cls_from_vci_and_mac(vci, mac):
    for cls in DEVICES:
        if cls.is_such_a_device(vci, mac):
            return cls
    return None     # not found

def register_all_devices():
    global DEVICES
    all_modules = []
    # add all modules in 'nodes' and 'switches' subdirs
    for package in (nodes, switches):
        all_modules.extend(module for name, module in \
                    inspect.getmembers(package, inspect.ismodule))
    for module in all_modules:
        DEVICES.extend(module.get_device_classes())

register_all_devices()




# auto-generated by dev/info-updater.py
# using metadata from dev/metadata.py

SETUP_INFO = dict(
    name = 'walt-common',
    version = '4',
    install_requires = ['plumbum>=1.4.2'],
    author = 'WalT developers',
    author_email = 'walt-contact@univ-grenoble-alpes.fr',
    keywords = 'WalT testbed',
    license = '3-Clause BSD',
    url = 'https://walt-project.liglab.fr',
    description = 'WalT common python modules.'
)



#!/usr/bin/env python
import os, sys
from select import select
from socket import socket, error as SocketError
from walt.common.constants import WALT_SERVER_DAEMON_PORT
from walt.common.reusable import reusable
from walt.common.tools import BusyIndicator
from walt.common.tcp import Requests

# exceptions may occur if the client disconnects.
# we should ignore those.
class APIChannel(object):
    def __init__(self, sock_file):
        self.sock_file = sock_file
    def write(self, *args):
        if self.sock_file.closed:
            return
        self.sock_file.write(repr(args) + '\n')
    def read(self):
        if self.sock_file.closed:
            return None
        try:
            return eval(self.sock_file.readline())
        except (EOFError, SyntaxError, OSError, SocketError):
            return None
    def fileno(self):
        return self.sock_file.fileno()

# The following pair of classes allows to pass API calls efficiently
# over a network socket.
#
# On the emitter end, we will have the following kind of code:
#
# api = AttrCallAggregator(api_forwarding_func)
# ...
# res = api.my.super.function(1, 2, a=3)
#
# This will cause the following to be called:
# api_forwarding_func('my.super.function', [1, 2], {'a':3})
#
# then it is easy for api_backend_func to pass its arguments
# on the socket connection.
#
# on the other end, we have have a code like:
#
# api_call_runner = AttrCallRunner(api_handler)
# ...
# api_call_runner.do('my.super.function', [1, 2], {'a':3})
#
# This will call the following to be called:
# api_handler.my.super.function(1, 2, a=3)
#
# This mechanism is efficient because it sends the whole attribute path
# at once on the network.
# Drawback: It works with remote function calls only, i.e. you cannot
# read a remote attribute. Thus all remote attributes must be functions.

class AttrCallAggregator(object):
    def __init__(self, handler, path = ()):
        self.path = path
        self.handler = handler
    def __getattr__(self, attr):
        return AttrCallAggregator(self.handler, self.path + (attr,))
    def __call__(self, *args, **kwargs):
        path = '.'.join(self.path)
        return self.handler(path, args, kwargs)

class AttrCallRunner(object):
    def __init__(self, handler):
        self.handler = handler
    def do(self, path, args, kwargs):
        obj = self.handler
        for attr in path.split('.'):
            obj = getattr(obj, attr)
        return obj(*args, **kwargs)

# Sometimes we start a long running process on the server and wait for
# its completion. In this case, while we are waiting, the server may
# send api calls to update the client stdout or stderr.
# This object provide methods to wait() and still be able to accept
# these calls.
MAX_BLOCKING_TIME = 0.1

class Fake(object):
    def __getattr__(self, attr):
        return lambda: None
    def set_label(self, label):
        pass
    def set_default_label(self):
        pass

class LinkException(Exception):
    pass

DEFAULT_BUSY_LABEL = 'Server is working'

@reusable
class ServerAPIConnection(object):
    def __init__(self, server_ip, local_service, target_api):
        self.target_api = target_api
        self.server_ip = server_ip
        self.sock = socket()
        self.sock_file = self.sock.makefile('r+',0)
        self.api_channel = APIChannel(self.sock_file)
        self.remote_version = None
        is_interactive = os.isatty(sys.stdout.fileno()) and \
                            os.isatty(sys.stdin.fileno())
        self.client_proxy = AttrCallAggregator(self.handle_client_call)
        self.local_api_handler = AttrCallRunner(local_service)
        if is_interactive:
            self.indicator = BusyIndicator(DEFAULT_BUSY_LABEL)
        else:
            self.indicator = Fake()
        self.connected = False
    # since the object is reusable we must ensure we connect only once.
    def connect(self):
        if not self.connected:
            self.sock.connect((self.server_ip, WALT_SERVER_DAEMON_PORT))
            self.sock_file.write('%d\n%s\n' % (Requests.REQ_API_SESSION, self.target_api))
            self.remote_version = int(self.sock_file.readline().strip())
            self.connected = True
    def set_busy_label(self, label):
        self.indicator.set_label(label)
    def set_default_busy_label(self):
        self.indicator.set_default_label()
    def get_remote_version(self):
        return self.remote_version
    def handle_client_call(self, path, args, kwargs):
        if hasattr(self, path):
            # this is something implemented locally
            return getattr(self, path)(*args, **kwargs)
        else:
            # this is a remote api call
            return self.do_remote_api_call(path, args, kwargs)
    def do_remote_api_call(self, path, args, kwargs):
        # send the api call
        self.api_channel.write('API_CALL', path, args, kwargs)
        # wait for the result
        return self.wait_api_result()
    def handle_api_call(self, path, args, kwargs):
        res = self.local_api_handler.do(path, args, kwargs)
        self.api_channel.write('RESULT', res)
    def wait_api_result(self):
        self.indicator.start()
        api_result = None
        while True:
            self.indicator.update()
            timeout = MAX_BLOCKING_TIME
            r, w, e = select((self.sock,), (), (self.sock,), timeout)
            if len(e) > 0:  # if error
                break
            if len(r) == 0: # if timeout
                continue
            # otherwise, there is something to process on the api connection
            self.indicator.reset()  # apparently the server is no longer busy
            event = self.api_channel.read()
            if event != None:
                if event[0] == 'API_CALL':
                    self.handle_api_call(*event[1:])
                    continue
                elif event[0] == 'EXCEPTION':
                    print 'Unexpected server-side issue! %s' % event[1]
                    break
                elif event[0] == 'RESULT':
                    api_result = event[1]
                    break
            raise LinkException('Unexpected communication issue with the server.')
        self.indicator.done()
        return api_result
    def __del__(self):
        self.sock.close()

class VoidAPIService(object):
    pass

# This class provides a 'with' environment to connect to
# the server API.
class ServerAPILink(object):
    def __init__(self, server_ip, target_api, local_service = None):
        if local_service == None:
            local_service = VoidAPIService()
        self.conn = ServerAPIConnection(
            server_ip,
            local_service,
            target_api)
    def __enter__(self):
        self.conn.connect()
        return self.conn.client_proxy
    def __exit__(self, type, value, traceback):
        # do not close the connection right now, it might be reused
        # thanks to the @reusable decorator of ServerAPIConnection
        pass




import sys, signal, itertools
from multiprocessing import Pipe
from threading import Thread, current_thread
from select import select
from walt.common.evloop import EventLoop
from walt.common.tools import AutoCleaner, SimpleContainer
from walt.common.apilink import AttrCallAggregator, AttrCallRunner

class DownwardPropagatedException(Exception):
    pass

class EvThread(Thread):
    def __init__(self, manager, name):
        Thread.__init__(self, name = name)
        self.ev_loop = EventLoop()
        self.pipe_in, self.pipe_out = Pipe()
        self.ev_loop.register_listener(self)
        manager.register_thread(self)

    # mimic an event loop
    def __getattr__(self, attr):
        return getattr(self.ev_loop, attr)

    def run(self):
        try:
            with AutoCleaner(self):
                self.prepare()
                self.ev_loop.loop()
        except DownwardPropagatedException:
            print self.name + ' stopped because of propagated exception.'
        except BaseException as e:
            try:
                self.pipe_in.send(e) # propagate upward
            except BaseException as issue:
                self.pipe_in.send(Exception(str(e)))
            raise

    def fileno(self):
        return self.pipe_in.fileno()

    def forward_exception(self, e):
        self.pipe_out.send(e)

    def handle_event(self, ts):
        self.down_exception = self.pipe_in.recv()
        raise DownwardPropagatedException

    def prepare(self):
        pass    # optionally override in subclass

    def cleanup(self):
        pass    # optionally override in subclass

class EvThreadsManager(object):
    def __init__(self):
        self.threads = []
    def register_thread(self, t):
        self.threads.append(t)
    def start(self):
        try:
            for t in self.threads:
                t.start()
            read_set = tuple(t.pipe_out for t in self.threads)
            r, w, e = select(read_set, (), read_set)
            e = r[0].recv()
            raise e
        except BaseException as e:
            # propagate downward to threads that are still alive
            for t in self.threads:
                try:
                    t.forward_exception(e)
                except:
                    pass

WAIT_MAX = 0.1

class ThreadConnector:
    def connect(self, remote):
        self.pipe, remote.pipe = Pipe()
    def close(self):
        self.pipe.close()
    def fileno(self):
        return self.pipe.fileno()
    def write(self, obj):
        self.pipe.send(obj)
    def read(self):
        return self.pipe.recv()
    def poll(self):
        return self.pipe.poll()
    def wait_next_event(self):
        # we implement a loop in order to make this wait
        # interruptible (every WAIT_MAX seconds)
        while not self.pipe.poll(WAIT_MAX):
            pass

PRIORITIES = { 'RESULT':0, 'EXCEPTION':1, 'API_CALL':2 }

class RPCSession(object):
    def __init__(self, connector, remote_req_id, local_service):
        self.connector = connector
        self.async = AttrCallAggregator(self.async_runner)
        self.sync = AttrCallAggregator(self.sync_runner)
        self.remote_req_id = remote_req_id
        self.local_service = local_service
    def async_runner(self, path, args, kwargs):
        return self.connector.async_runner(
                self.remote_req_id, self.local_service, path, args, kwargs)
    def sync_runner(self, path, args, kwargs):
        return self.connector.sync_runner(
                self.remote_req_id, self.local_service, path, args, kwargs)

class RPCTask(object):
    def __init__(self, connector, remote_req_id):
        self.connector = connector
        self.remote_req_id = remote_req_id
    def return_result(self, res):
        #print current_thread().name, 'RESULT', self.remote_req_id, res
        self.connector.write(('RESULT', self.remote_req_id, res))
    def return_exception(self, e):
        print current_thread().name + ': Exception occured while performing API request:'
        sys.excepthook(*sys.exc_info())
        self.connector.write(('RESULT', self.remote_req_id, Exception(str(e))))

class RPCContext(object):
    def __init__(self, connector, remote_req_id, local_service):
        self.requester = RPCSession(connector, remote_req_id, local_service)
        self.task = RPCTask(connector, remote_req_id)

class RPCThreadConnector(ThreadConnector):
    def __init__(self, default_service = None):
        self.submitted_tasks = {}
        self.ids_generator = itertools.count()
        self.results = {}
        self.default_service = AttrCallRunner(default_service)
        self.default_session = self.create_session(-1, default_service)
        self.async = self.default_session.async
        self.sync = self.default_session.sync
    def local_service(self, local_service):
        return self.create_session(-1, local_service)
    def create_session(self, remote_req_id, local_service):
        return RPCSession(self, remote_req_id, local_service)
    def handle_event(self, ts):
        return self.handle_next_event()
    def handle_next_event(self):
        events = []
        while self.poll():
            events.append(self.read())
        if len(events) == 0:
            return False    # no data, quit
        events.sort(key=lambda x: PRIORITIES[x[0]])
        for event in events:
            if event[0] == 'API_CALL':
                self.handle_api_call(*event[1:])
                continue
            elif event[0] == 'RESULT':
                local_req_id, result = event[1], event[2]
                sync = self.submitted_tasks[local_req_id].sync
                cb = self.submitted_tasks[local_req_id].result_cb
                if cb != None:
                    cb(result)
                if sync:
                    self.results[local_req_id] = result
                del self.submitted_tasks[local_req_id]
                continue
            raise Exception('Broken communication with remote end.')
    def handle_api_call(self, local_req_id, remote_req_id, path, args, kwargs, sync):
        if local_req_id == -1:
            local_service = self.default_service
        else:
            local_service = self.submitted_tasks[local_req_id].local_service
        context = RPCContext(self, remote_req_id, local_service)
        args = (context,) + args
        try:
            res = local_service.do(path, args, kwargs)
        except BaseException as e:
            context.task.return_exception(e)
            return
        if sync:
            context.task.return_result(res)
    def then(self, cb): # specify callback
        self.submitted_tasks[self.last_req_id].result_cb = cb
    def async_runner(self, remote_req_id, local_service, path, args, kwargs):
        local_req_id = self.send_task(remote_req_id, local_service, path, args, kwargs, False)
        return self
    def sync_runner(self, remote_req_id, local_service, path, args, kwargs):
        local_req_id = self.send_task(remote_req_id, local_service, path, args, kwargs, True)
        while not local_req_id in self.results:
            self.wait_next_event()
            status = self.handle_next_event()
            if status == False:
                return None
        return self.results.pop(local_req_id)
    def send_task(self, remote_req_id, local_service, path, args, kwargs, sync):
        local_req_id = self.ids_generator.next()
        self.last_req_id = local_req_id
        self.submitted_tasks[local_req_id] = SimpleContainer(
                    local_service = AttrCallRunner(local_service),
                    path = path,
                    args = args,
                    kwargs = kwargs,
                    sync = sync,
                    result_cb = None)
        #print current_thread().name, 'API_CALL', remote_req_id, local_req_id, path, args, kwargs, sync
        self.write(('API_CALL', remote_req_id, local_req_id, path, args, kwargs, sync))
        return local_req_id




class Fake(object):
    MAC_PREFIX = "xx:xx:xx"
    MODEL_NAME = "fake"
    WALT_TYPE  = "node"
    @staticmethod
    def is_such_a_device(vci, mac):
        return False

def get_device_classes():
    return (Fake,)





RPI_MAC_PREFIX = "b8:27:eb"

def create_rpi_class(cls_name, model):
    # class static methods
    def is_such_a_device(vci, mac):
        return vci == model
    # class creation
    return type(cls_name, (), dict(
        MAC_PREFIX = RPI_MAC_PREFIX,
        MODEL_NAME = model,
        WALT_TYPE  = "node",
        is_such_a_device = staticmethod(is_such_a_device)
    ))
    
# creation of one class for each model
RpiB = create_rpi_class('RpiB', 'rpi-b')
RpiBPlus = create_rpi_class('RpiBPlus', 'rpi-b-plus')
Rpi2B = create_rpi_class('Rpi2B', 'rpi-2-b')
Rpi3B = create_rpi_class('Rpi3B', 'rpi-3-b')

def get_device_classes():
    return (RpiB, RpiBPlus, Rpi2B, Rpi3B)




import subprocess, os, sys, json, re, signal
from plumbum.cmd import cat
from collections import OrderedDict
from datetime import datetime, timedelta
from fcntl import fcntl, F_GETFD, F_SETFD, F_GETFL, F_SETFL, FD_CLOEXEC

DEVNULL = open(os.devnull, 'w')

def eval_cmd(cmd):
    return cmd()

def get_mac_address(interface):
    return eval_cmd(cat["/sys/class/net/" + interface + "/address"]).strip()

def do(cmd):
    return subprocess.call(cmd, stdout=DEVNULL, shell=True)

def succeeds(cmd):
    return do(cmd) == 0

def failsafe_makedirs(path):
    # remove if not a dir
    if os.path.lexists(path) and not os.path.isdir(path):
        os.remove(path)
    # create only if missing
    if not os.path.exists(path):
        os.makedirs(path)

def failsafe_symlink(src_target, dst_path, force_relative = False):
    # if force_relative, turn src_target into a relative path
    if force_relative:
        targetwords = src_target.rstrip('/').split('/')
        pathwords = os.path.dirname(dst_path).split('/')
        num_common = len([x for x, y in zip(targetwords, pathwords) if x == y])
        relativewords = [ '..' ] * (len(pathwords) - num_common) + targetwords[num_common:]
        src_target = '/'.join(relativewords)
    # remove existing symlink if any
    if os.path.lexists(dst_path):
        if os.readlink(dst_path) == src_target:
            # nothing to do
            return
        else:
            # symlink target (src_target) has changed
            os.remove(dst_path)
    # ensure parent dir of dst_path exists
    failsafe_makedirs(os.path.dirname(dst_path))
    # create the symlink
    os.symlink(src_target, dst_path)

# Note: json comments are not allowed in the standard
# and thus not handled in the json python module.
# We handle them manually.
def read_json(file_path):
    content = None
    try:
        with open(file_path) as f:
            # filter out comments
            filtered = re.sub('#.*', '', f.read())
            # read valid json
            content = json.loads(filtered, object_pairs_hook=OrderedDict)
    except:
        pass
    return content

# use the following like this:
#
# with AutoCleaner(<obj>) as <var>:
#     ... work_with <var> ...
#
# <obj> must provide a method cleanup()
# that will be called automatically when leaving
# the with construct.

class AutoCleaner(object):
    def __init__(self, obj):
        self.obj = obj
    def __enter__(self):
        return self.obj
    def __exit__(self, t, value, traceback):
        self.obj.cleanup()
        self.obj = None

# look for a kernel parameter in /proc/cmdline
# if in the form <arg>=<val> return <val>
# if in the form <arg> return True
# if not found return None

def get_kernel_bootarg(in_bootarg):
    with open('/proc/cmdline') as f:
        for bootarg in f.read().split():
            name, val = (bootarg.split('=') + [ True ])[:2]
            if name == in_bootarg:
                return val

PROGRESS_INDICATOR_PERIOD = timedelta(seconds=1.0)

class BusyIndicator(object):
    def __init__(self, label):
        self.default_label = label
        self.label = label
        self.last_time = None
        self.next_time = None
        self.msg_len = None
    def start(self):
        self.last_time = None
        self.next_time = datetime.now() + PROGRESS_INDICATOR_PERIOD
    def write_stdout(self, s):
        sys.stdout.write(s)
        sys.stdout.flush()
    def update(self):
        if datetime.now() > self.next_time:
            if self.last_time == None:
                msg = self.label + "... *"
                self.write_stdout(msg)
                self.msg_len = len(msg)
            else:
                self.write_stdout("*")
                self.msg_len += 1
            self.last_time = datetime.now()
            self.next_time = self.last_time + PROGRESS_INDICATOR_PERIOD
    def done(self):
        if self.last_time != None:
            self.write_stdout("\r" + (' ' * self.msg_len) + "\r")
    def reset(self):
        self.done()
        self.start()
    def set_label(self, label):
        self.label = label
        if self.last_time != None:
            self.reset()
    def set_default_label(self):
        self.set_label(self.default_label)

def fd_copy(fd_src, fd_dst, size):
    try:
        s = os.read(fd_src, size)
        if len(s) == 0:
            return None
        os.write(fd_dst, s)
        return s
    except:
        return None

def set_non_blocking(fd):
    fl = fcntl(fd, F_GETFL)
    fcntl(fd, F_SETFL, fl | os.O_NONBLOCK)

def remove_non_utf8(s):
    return s.decode('utf-8','ignore').encode("utf-8")

def on_sigterm_throw_exception():
    def signal_handler(signal, frame):
        print('SIGTERM received.')
        raise KeyboardInterrupt
    signal.signal(signal.SIGTERM, signal_handler)

MAX_PRINTED_NODES = 10
CONJUGATE_REGEXP = r'\b(\w*)\((\w*)\)'

def format_sentence(sentence, items,
            label_none, label_singular, label_plural):
    # conjugate if plural or singular
    if len(items) == 1:
        sentence = re.sub(CONJUGATE_REGEXP, r'\1', sentence)
    else:
        sentence = re.sub(CONJUGATE_REGEXP, r'\2', sentence)
    # designation of items
    sorted_items = sorted(items)
    if len(items) > MAX_PRINTED_NODES:
        s_items = '%s %s, %s, ..., %s' % (label_plural, sorted_items[0], sorted_items[1], sorted_items[-1])
    elif len(items) == 0:
        s_items = label_none
    elif len(items) > 1:
        s_items = '%s %s and %s' % (label_plural, ', '.join(sorted_items[:-1]), sorted_items[-1])
    else:   # 1 item
        s_items = '%s %s' % (label_singular, sorted_items[0])
    # almost done
    return sentence % s_items

def format_sentence_about_nodes(sentence, nodes):
    """
    example 1:
        input: sentence = '%s seems(seem) dead.', nodes = ['rpi0']
        output: 'Node rpi0 seems dead.'
    example 2:
        input: sentence = '%s seems(seem) dead.', nodes = ['rpi0', 'rpi1', 'rpi2']
        output: 'Nodes rpi0, rpi1 and rpi2 seem dead.'
    (and if there are many nodes, an ellipsis is used.)
    """
    return format_sentence(sentence, nodes, 'No nodes', 'Node', 'Nodes')

class SimpleContainer(object):
    def __init__(self, **args):
        self.update(**args)
    def update(self, **args):
        self.__dict__.update(**args)
        return self
    def copy(self):
        return SimpleContainer(**self.__dict__)

def serialize_ordered_dict(od):
    res = []
    for k, v in od.items():
        if isinstance(v, OrderedDict):
            v = serialize_ordered_dict(v)
        res.append((k, v))
    return tuple(res)

def deserialize_ordered_dict(t):
    d = OrderedDict()
    for k, v in t:
        if isinstance(v, tuple):
            v = deserialize_ordered_dict(v)
        d[k] = v
    return d

def set_close_on_exec(fd, close_on_exec):
    val = fcntl(fd, F_GETFD)
    if close_on_exec:
        val |= FD_CLOEXEC
    else:
        val &= ~FD_CLOEXEC
    fcntl(fd, F_SETFD, val)

def restart():
    os.execvp(sys.argv[0], sys.argv)



from setuptools import setup, find_packages
import sys

if not sys.version_info[0] == 2:
    sys.exit("Sorry, Python 3 is not supported (yet)")

# make sure we are executing the new info.py
execfile('walt/common/info.py')
SETUP_INFO.update(
    packages = find_packages()
)
setup(**SETUP_INFO)



# cryptography design notes
# -------------------------
# we want to keep the installation of walt-client simple
# and portable (pip install walt-client), which forbids the
# use of a python module relying on external libraries,
# such as pycrypto, cryptography, etc. (they require
# python-dev to be installed).
#
# Our need in terms of security is very lightweight:
# * we use encrypted data very rarely (case of 'walt image publish' only)
# * we encrypt a very small amount of data (a password)
#
# Given these facts, we have implemented a simple process:
# At each 'walt image publish' command:
# 1) server and client generate a random key
# 2) an encryption key is agreed using Diffie Hellman process
# 3) the password is encrypted based on this encryption key
#    using a pure-python implementation of the blowfish algorithm.




NETGEAR_MAC_PREFIX = "6c:b0:ce"
NETGEAR_MAC_PREFIXES = [ "6c:b0:ce", "28:c6:8e", "44:94:fc" ]

class NetgearSwitch(object):
    MAC_PREFIX = NETGEAR_MAC_PREFIX
    MODEL_NAME = "netgear"
    WALT_TYPE  = "switch"
    @staticmethod
    def is_such_a_device(vci, mac):
        #return mac.startswith(NETGEAR_MAC_PREFIX)
        return (len(filter(mac.startswith,NETGEAR_MAC_PREFIXES+[''])[0]) != 0)

def get_device_classes():
    return (NetgearSwitch,)




import os, termios, struct, fcntl, re

TTY_CONTROL_CHARS = ''.join(map(unichr, range(0,32) + range(127,160)))
TTY_CONTROL_CHARS_RE = re.compile('[%s]' % re.escape(TTY_CONTROL_CHARS))

def remove_control_chars(s):
    return TTY_CONTROL_CHARS_RE.sub('', s)

def set_tty_size_raw(fd, tty_size_raw):
    fcntl.ioctl(fd, termios.TIOCSWINSZ, tty_size_raw)

def set_tty_size(fd, tty_size):
    tty_rows, tty_cols = tty_size
    packed = struct.pack("HHHH", tty_rows, tty_cols, 0, 0)
    set_tty_size_raw(fd, packed)

def acquire_controlling_tty(tty_fd):
    # we need to be a session leader
    os.setsid()
    # just opening the file descriptor *without* the flag O_NOCTTY
    # should be enough
    ttyname = os.ttyname(tty_fd)
    os.close(os.open(ttyname, os.O_RDONLY))

def tty_disable_echoctl(fd):
    attrs = termios.tcgetattr(fd)
    attrs[3] = attrs[3] & ~termios.ECHOCTL
    termios.tcsetattr(fd, termios.TCSADRAIN, attrs)




# -*- coding: utf-8 -*-
explorer = None

def register_api_explorer(expl):
    global explorer
    explorer = expl

# mark methods to be exposed by replacing them by
# an instance of APIExposedMethod class.
# (we cannot rename these attributes here, their 'name'
# belongs to the class. That's why we need
# the other decorator @api below.)
class APIExposedMethod(object):
    def __init__(self, func):
        self.func = func

def api_expose_method(func):
    return APIExposedMethod(func)

# given an __init__ constructor and a set of
# attributes that must be exposed, compute a decorated
# __init__ function, that will rename the attrs at the
# end of the provided constructor.
class APIDecoratedInit(object):
    def __init__(self, attrs, init_func):
        self.attrs = attrs
        def decorated_init_func(obj, *args, **kwargs):
            init_func(obj, *args, **kwargs)
            for attr in attrs:
                val = getattr(obj, attr)
                setattr(obj, 'exposed_' + attr, val)
        self.decorated_init = decorated_init_func

# return a decorator to be applied to an __init__ function
# and able to handle a list of attributes that must be exposed.
def api_expose_attrs(*attrs):
    def decorator(init_func):
        return APIDecoratedInit(attrs, init_func)
    return decorator

# expose class attributes specified.
def api_expose_class_attrs(*c_attrs):
    def decorator(cls):
        for c_attr in c_attrs:
            if explorer:
                explorer.add_attr(cls, c_attr)
            val = getattr(cls, c_attr)
            setattr(cls, 'exposed_' + c_attr, val)
        return cls
    return decorator

# Behavior in production code:
# loop through attributes of the class...
# * look for any instance of class APIExposedMethod,
# replace these by the original method, and
# duplicate them by prefixing 'exposed_'
# * look for any instance of class APIDecoratedInit,
# replace these by the decorated __init__
#
# Behavior when exploring the API:
# dump the prototype of the exposed methods and
# the name of the exposed attributes.
def api(cls):
    for k, v in list(cls.__dict__.items()):
        if isinstance(v, APIExposedMethod):
            if explorer:
                explorer.explore_method(cls, v.func)
            setattr(cls, k, v.func)
            setattr(cls, "exposed_%s" % k, v.func)
        elif isinstance(v, APIDecoratedInit):
            if explorer:
                for attr in v.attrs:
                    explorer.add_attr(cls, attr)
            setattr(cls, k, v.decorated_init)
    return cls




#!/usr/bin/env python
import os, stat, select
from walt.common.tools import failsafe_makedirs, fd_copy
from threading import Thread

RW_ALL = stat.S_IRUSR | stat.S_IWUSR | \
         stat.S_IRGRP | stat.S_IWGRP | \
         stat.S_IROTH | stat.S_IWOTH
BUFFER_SIZE = 1024

def failsafe_mkfifo(path, mode = None):
    if mode == None:
        mode = RW_ALL
    # check if it does not exist already
    if os.path.exists(path):
        return
    # ensure parent dir exists
    failsafe_makedirs(os.path.dirname(path))
    # create the fifo
    os.mkfifo(path)
    os.chmod(path, mode)

# fifos are special files.
# * when opening on one end, it will block until another process attempts to
#   open the other end.
# * when closed on one end, we should also close on the other end. For example,
#   if we are reading on /tmp/fifo and issuie 'echo test > /tmp/fifo' on another
#   terminal, we will receive 'test' but then we cannot start reading again
#   because the fifo has been closed on the 'echo' side. Instead, we should
#   close and re-open the fifo in order to get the next request.
# Because of that, we handle the fifo objects by running a dedicated thread
# that releases the main thread from fifo management work. The dedicated thread
# just dumps everything from the fifo to a pipe. The main thread reads the other
# end of this pipe. This allows the main thread to implement usual IO mechanisms
# (select(), etc.).

class ReadableFifo(object):
    def __init__(self, path):
        self.path = path
    def open(self):
        failsafe_mkfifo(self.path)
        data_r, data_w = os.pipe()
        ctrl_r, self.ctrl_w = os.pipe()
        t = Thread(target=self.reader_thread, args=(data_w, ctrl_r))
        t.start()
        self.file_r = os.fdopen(data_r, 'r', 0)
    def reader_thread(self, data_w, ctrl_r):
        should_stop = False
        while not should_stop:
            fifo_fd = os.open(self.path, os.O_RDONLY | os.O_NONBLOCK)
            fds = [fifo_fd, ctrl_r]
            while len(fds) == 2:
                r, w, e = select.select(fds,[],[])
                if fifo_fd in r:
                    fd_copy(fifo_fd, data_w, BUFFER_SIZE) or fds.remove(fifo_fd)
                else:
                    os.read(ctrl_r, 1)
                    should_stop = True
                    break
            os.close(fifo_fd)
        os.close(data_w)
        os.close(ctrl_r)
    def close(self):
        try:
            os.write(self.ctrl_w, 'X')
            os.close(self.ctrl_w)
        except:
            pass
        try:
            self.file_r.close()
        except:
            pass
    def __enter__(self):
        return self.file_r
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
    # redirect other accesses to self.file_r
    def __getattr__(self, attr):
        return getattr(self.file_r, attr)

def open_readable_fifo(path):
    fifo = ReadableFifo(path)
    fifo.open()
    return fifo




#!/usr/bin/env python
import random

class DiffieHellman(object):
    BITSIZE = 2048

    # RFC 5114 section 2.3
    p = int(''.join("""
    87A8E61D B4B6663C FFBBD19C 65195999 8CEEF608 660DD0F2
    5D2CEED4 435E3B00 E00DF8F1 D61957D4 FAF7DF45 61B2AA30
    16C3D911 34096FAA 3BF4296D 830E9A7C 209E0C64 97517ABD
    5A8A9D30 6BCF67ED 91F9E672 5B4758C0 22E0B1EF 4275BF7B
    6C5BFC11 D45F9088 B941F54E B1E59BB8 BC39A0BF 12307F5C
    4FDB70C5 81B23F76 B63ACAE1 CAA6B790 2D525267 35488A0E
    F13C6D9A 51BFA4AB 3AD83477 96524D8E F6A167B5 A41825D9
    67E144E5 14056425 1CCACB83 E6B486F6 B3CA3F79 71506026
    C0B857F6 89962856 DED4010A BD0BE621 C3A3960A 54E710C3
    75F26375 D7014103 A4B54330 C198AF12 6116D227 6E11715F
    693877FA D7EF09CA DB094AE9 1E1A1597
    """.split()), 16)

    g = int(''.join("""
    3FB32C9B 73134D0B 2E775066 60EDBD48 4CA7B18F 21EF2054
    07F4793A 1A0BA125 10DBC150 77BE463F FF4FED4A AC0BB555
    BE3A6C1B 0C6B47B1 BC3773BF 7E8C6F62 901228F8 C28CBB18
    A55AE313 41000A65 0196F931 C77A57F2 DDF463E5 E9EC144B
    777DE62A AAB8A862 8AC376D2 82D6ED38 64E67982 428EBC83
    1D14348F 6F2F9193 B5045AF2 767164E1 DFC967C1 FB3F2E55
    A4BD1BFF E83B9C80 D052B985 D182EA0A DB2A3B73 13D3FE14
    C8484B1E 052588B9 B7D2BBD2 DF016199 ECD06E15 57CD0915
    B3353BBB 64E0EC37 7FD02837 0DF92B52 C7891428 CDC67EB6
    184B523D 1DB246C3 2F630784 90F00EF8 D647D148 D4795451
    5E2327CF EF98C582 664B4C0F 6CC41659
    """.split()), 16)

    @staticmethod
    def generate_priv_key():
        return random.getrandbits(DiffieHellman.BITSIZE)

    @staticmethod
    def generate_pub_key(priv_key):
        return pow(DiffieHellman.g, priv_key, DiffieHellman.p)

    @staticmethod
    def generate_symmetric_key(remote_pub_key, priv_key):
        return pow(remote_pub_key, priv_key, DiffieHellman.p)

class DHPeer(object):
    def __init__(self):
        self.priv_key = DiffieHellman.generate_priv_key()
        self.pub_key = DiffieHellman.generate_pub_key(self.priv_key)
        self.symmetric_key = None
    def get_pub_key(self):
        return self.pub_key
    def establish_session(self, remote_pub_key):
        self.symmetric_key = DiffieHellman.generate_symmetric_key(
                    remote_pub_key, self.priv_key)

